/* -*-c-*-
 *
 * $Id: mpmont.c,v 1.14 2001/02/22 09:04:26 mdw Exp $
 *
 * Montgomery reduction
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mpmont.c,v $
 * Revision 1.14  2001/02/22 09:04:26  mdw
 * Cosmetic fix.
 *
 * Revision 1.13  2001/02/03 12:00:29  mdw
 * Now @mp_drop@ checks its argument is non-NULL before attempting to free
 * it.  Note that the macro version @MP_DROP@ doesn't do this.
 *
 * Revision 1.12  2000/10/08 15:48:35  mdw
 * Rename Karatsuba constants now that we have @gfx_kmul@ too.
 *
 * Revision 1.11  2000/10/08 12:04:27  mdw
 * (mpmont_reduce, mpmont_mul): Cope with negative numbers.
 *
 * Revision 1.10  2000/07/29 17:05:43  mdw
 * (mpmont_expr): Use sliding window exponentiation, with a drop-through
 * for small exponents to use a simple left-to-right bitwise routine.  This
 * can reduce modexp times by up to a quarter.
 *
 * Revision 1.9  2000/06/17 11:45:09  mdw
 * Major memory management overhaul.  Added arena support.  Use the secure
 * arena for secret integers.  Replace and improve the MP management macros
 * (e.g., replace MP_MODIFY by MP_DEST).
 *
 * Revision 1.8  1999/12/22 15:55:00  mdw
 * Adjust Karatsuba parameters.
 *
 * Revision 1.7  1999/12/11 01:51:14  mdw
 * Use a Karatsuba-based reduction for large moduli.
 *
 * Revision 1.6  1999/12/10 23:18:39  mdw
 * Change interface for suggested destinations.
 *
 * Revision 1.5  1999/11/22 13:58:40  mdw
 * Add an option to disable Montgomery reduction, so that performance
 * comparisons can be done.
 *
 * Revision 1.4  1999/11/21 12:27:06  mdw
 * Remove a division from the Montgomery setup by calculating
 * %$R^2 \bmod m$% first and then %$R \bmod m$% by Montgomery reduction of
 * %$R^2$%.
 *
 * Revision 1.3  1999/11/21 11:35:10  mdw
 * Performance improvement: use @mp_sqr@ and @mpmont_reduce@ instead of
 * @mpmont_mul@ for squaring in exponentiation.
 *
 * Revision 1.2  1999/11/19 13:17:26  mdw
 * Add extra interface to exponentiation which returns a Montgomerized
 * result.
 *
 * Revision 1.1  1999/11/17 18:02:16  mdw
 * New multiprecision integer arithmetic suite.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "mp.h"
#include "mpmont.h"

/*----- Tweakables --------------------------------------------------------*/

/* --- @MPMONT_DISABLE@ --- *
 *
 * Replace all the clever Montgomery reduction with good old-fashioned long
 * division.
 */

/* #define MPMONT_DISABLE */

/*----- Main code ---------------------------------------------------------*/

/* --- @mpmont_create@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *m@ = modulus to use
 *
 * Returns:	---
 *
 * Use:		Initializes a Montgomery reduction context ready for use.
 *		The argument @m@ must be a positive odd integer.
 */

#ifdef MPMONT_DISABLE

void mpmont_create(mpmont *mm, mp *m)
{
  mp_shrink(m);
  mm->m = MP_COPY(m);
  mm->r = MP_ONE;
  mm->r2 = MP_ONE;
  mm->mi = MP_ONE;
}

#else

void mpmont_create(mpmont *mm, mp *m)
{
  size_t n = MP_LEN(m);
  mp *r2 = mp_new(2 * n + 1, 0);
  mp r;

  /* --- Validate the arguments --- */

  assert(((void)"Montgomery modulus must be positive",
	  (m->f & MP_NEG) == 0));
  assert(((void)"Montgomery modulus must be odd", m->v[0] & 1));

  /* --- Take a copy of the modulus --- */

  mp_shrink(m);
  mm->m = MP_COPY(m);

  /* --- Determine %$R^2$% --- */

  mm->n = n;
  MPX_ZERO(r2->v, r2->vl - 1);
  r2->vl[-1] = 1;

  /* --- Find the magic value @mi@ --- */

  mp_build(&r, r2->v + n, r2->vl);
  mm->mi = MP_NEW;
  mp_gcd(0, 0, &mm->mi, &r, m);
  mm->mi = mp_sub(mm->mi, &r, mm->mi);

  /* --- Discover the values %$R \bmod m$% and %$R^2 \bmod m$% --- */

  mm->r2 = MP_NEW;
  mp_div(0, &mm->r2, r2, m);
  mm->r = mpmont_reduce(mm, MP_NEW, mm->r2);
  MP_DROP(r2);
}

#endif

/* --- @mpmont_destroy@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to a Montgomery reduction context
 *
 * Returns:	---
 *
 * Use:		Disposes of a context when it's no longer of any use to
 *		anyone.
 */

void mpmont_destroy(mpmont *mm)
{
  MP_DROP(mm->m);
  MP_DROP(mm->r);
  MP_DROP(mm->r2);
  MP_DROP(mm->mi);
}

/* --- @mpmont_reduce@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@mp *a@ = source, assumed positive
 *
 * Returns:	Result, %$a R^{-1} \bmod m$%.
 */

#ifdef MPMONT_DISABLE

mp *mpmont_reduce(mpmont *mm, mp *d, mp *a)
{
  mp_div(0, &d, a, mm->m);
  return (d);
}

#else

mp *mpmont_reduce(mpmont *mm, mp *d, mp *a)
{
  size_t n = mm->n;

  /* --- Check for serious Karatsuba reduction --- */

  if (n > MPK_THRESH * 3) {
    mp al;
    mpw *vl;
    mp *u;

    if (MP_LEN(a) >= n)
      vl = a->v + n;
    else
      vl = a->vl;
    mp_build(&al, a->v, vl);
    u = mp_mul(MP_NEW, &al, mm->mi);
    if (MP_LEN(u) > n)
      u->vl = u->v + n;
    u = mp_mul(u, u, mm->m);
    d = mp_add(d, a, u);
    mp_drop(u);
  }

  /* --- Otherwise do it the hard way --- */

  else {
    mpw *dv, *dvl;
    mpw *mv, *mvl;
    mpw mi;
    size_t k = n;

    /* --- Initial conditioning of the arguments --- */

    a = MP_COPY(a);
    if (d)
      MP_DROP(d);
    d = a;
    MP_DEST(d, 2 * n + 1, a->f);

    dv = d->v; dvl = d->vl;
    mv = mm->m->v; mvl = mm->m->vl;

    /* --- Let's go to work --- */

    mi = mm->mi->v[0];
    while (k--) {
      mpw u = MPW(*dv * mi);
      MPX_UMLAN(dv, dvl, mv, mvl, u);
      dv++;
    }
  }

  /* --- Wrap everything up --- */

  memmove(d->v, d->v + n, MPWS(MP_LEN(d) - n));
  d->vl -= n;
  if (MPX_UCMP(d->v, d->vl, >=, mm->m->v, mm->m->vl))
    mpx_usub(d->v, d->vl, d->v, d->vl, mm->m->v, mm->m->vl);
  if (d->f & MP_NEG) {
    mpx_usub(d->v, d->vl, mm->m->v, mm->m->vl, d->v, d->vl);
    d->f &= ~MP_NEG;
  }
  MP_SHRINK(d);
  return (d);
}

#endif

/* --- @mpmont_mul@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = destination
 *		@mp *a, *b@ = sources, assumed positive
 *
 * Returns:	Result, %$a b R^{-1} \bmod m$%.
 */

#ifdef MPMONT_DISABLE

mp *mpmont_mul(mpmont *mm, mp *d, mp *a, mp *b)
{
  d = mp_mul(d, a, b);
  mp_div(0, &d, d, mm->m);
  return (d);
}

#else

mp *mpmont_mul(mpmont *mm, mp *d, mp *a, mp *b)
{
  if (mm->n > MPK_THRESH * 3) {
    d = mp_mul(d, a, b);
    d = mpmont_reduce(mm, d, d);
  } else {
    mpw *dv, *dvl;
    mpw *av, *avl;
    mpw *bv, *bvl;
    mpw *mv, *mvl;
    mpw y;
    size_t n, i;
    mpw mi;

    /* --- Initial conditioning of the arguments --- */

    if (MP_LEN(a) > MP_LEN(b)) {
      mp *t = a; a = b; b = t;
    }
    n = MP_LEN(mm->m);

    a = MP_COPY(a);
    b = MP_COPY(b);
    MP_DEST(d, 2 * n + 1, a->f | b->f | MP_UNDEF);
    dv = d->v; dvl = d->vl;
    MPX_ZERO(dv, dvl);
    av = a->v; avl = a->vl;
    bv = b->v; bvl = b->vl;
    mv = mm->m->v; mvl = mm->m->vl;
    y = *bv;

    /* --- Montgomery multiplication phase --- */

    i = 0;
    mi = mm->mi->v[0];
    while (i < n && av < avl) {
      mpw x = *av++;
      mpw u = MPW((*dv + x * y) * mi);
      MPX_UMLAN(dv, dvl, bv, bvl, x);
      MPX_UMLAN(dv, dvl, mv, mvl, u);
      dv++;
      i++;
    }

    /* --- Simpler Montgomery reduction phase --- */

    while (i < n) {
      mpw u = MPW(*dv * mi);
      MPX_UMLAN(dv, dvl, mv, mvl, u);
      dv++;
      i++;
    }

    /* --- Done --- */

    memmove(d->v, dv, MPWS(dvl - dv));
    d->vl -= dv - d->v;
    if (MPX_UCMP(d->v, d->vl, >=, mm->m->v, mm->m->vl))
      mpx_usub(d->v, d->vl, d->v, d->vl, mm->m->v, mm->m->vl);
    if ((a->f ^ b->f) & MP_NEG)
      mpx_usub(d->v, d->vl, mm->m->v, mm->m->vl, d->v, d->vl);
    MP_SHRINK(d);
    d->f = (a->f | b->f) & MP_BURN;
    MP_DROP(a);
    MP_DROP(b);
  }

  return (d);
}

#endif

/* --- @mpmont_expr@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@mp *a@ = base
 *		@mp *e@ = exponent
 *
 * Returns:	Result, %$a^e R \bmod m$%.
 */

#define WINSZ 5
#define TABSZ (1 << (WINSZ - 1))

#define THRESH (((MPW_BITS / WINSZ) << 2) + 1)

static mp *exp_simple(mpmont *mm, mp *d, mp *a, mp *e)
{
  mpscan sc;
  mp *ar;
  mp *x = MP_COPY(mm->r);
  mp *spare = (e->f & MP_BURN) ? MP_NEWSEC : MP_NEW;
  unsigned sq = 0;

  mp_rscan(&sc, e);
  if (!MP_RSTEP(&sc))
    goto exit;
  while (!MP_RBIT(&sc))
    MP_RSTEP(&sc);

  /* --- Do the main body of the work --- */
  
  ar = mpmont_mul(mm, MP_NEW, a, mm->r2);
  for (;;) {
    sq++;
    while (sq) {
      mp *y;
      y = mp_sqr(spare, x);
      y = mpmont_reduce(mm, y, y);
      spare = x; x = y;
      sq--;
    }
    { mp *y = mpmont_mul(mm, spare, x, ar); spare = x; x = y; }
    sq = 0;
    for (;;) {
      if (!MP_RSTEP(&sc))
	goto done;
      if (MP_RBIT(&sc))
	break;
      sq++;
    }
  }

  /* --- Do a final round of squaring --- */

done:
  while (sq) {
    mp *y;
    y = mp_sqr(spare, x);
    y = mpmont_reduce(mm, y, y);
    spare = x; x = y;
    sq--;
  }

  /* --- Done --- */

  MP_DROP(ar);
exit:
  if (spare != MP_NEW)
    MP_DROP(spare);
  if (d != MP_NEW)
    MP_DROP(d);
  return (x);
}

mp *mpmont_expr(mpmont *mm, mp *d, mp *a, mp *e)
{
  mp **tab;
  mp *ar, *a2;
  mp *spare = (e->f & MP_BURN) ? MP_NEWSEC : MP_NEW;
  mp *x = MP_COPY(mm->r);
  unsigned i, sq = 0;
  mpscan sc;

  /* --- Do we bother? --- */

  MP_SHRINK(e);
  if (MP_LEN(e) == 0)
    goto exit;
  if (MP_LEN(e) < THRESH) {
    x->ref--;
    return (exp_simple(mm, d, a, e));
  }

  /* --- Do the precomputation --- */

  ar = mpmont_mul(mm, MP_NEW, a, mm->r2);
  a2 = mp_sqr(MP_NEW, ar);
  a2 = mpmont_reduce(mm, a2, a2);
  tab = xmalloc(TABSZ * sizeof(mp *));
  tab[0] = ar;
  for (i = 1; i < TABSZ; i++)
    tab[i] = mpmont_mul(mm, MP_NEW, tab[i - 1], a2);
  mp_drop(a2);
  mp_rscan(&sc, e);

  /* --- Skip top-end zero bits --- *
   *
   * If the initial step worked, there must be a set bit somewhere, so keep
   * stepping until I find it.
   */

  MP_RSTEP(&sc);
  while (!MP_RBIT(&sc))
    MP_RSTEP(&sc);

  /* --- Now for the main work --- */

  for (;;) {
    unsigned l = 0;
    unsigned z = 0;

    /* --- The next bit is set, so read a window index --- *
     *
     * Reset @i@ to zero and increment @sq@.  Then, until either I read
     * @WINSZ@ bits or I run out of bits, scan in a bit: if it's clear, bump
     * the @z@ counter; if it's set, push a set bit into @i@, shift it over
     * by @z@ bits, bump @sq@ by @z + 1@ and clear @z@.  By the end of this
     * palaver, @i@ is an index to the precomputed value in @tab@.
     */

    i = 0;
    sq++;
    for (;;) {
      l++;
      if (l >= WINSZ || !MP_RSTEP(&sc))
	break;
      if (!MP_RBIT(&sc))
	z++;
      else {
	i = ((i << 1) | 1) << z;
	sq += z + 1;
	z = 0;
      }
    }

    /* --- Do the squaring --- *
     *
     * Remember that @sq@ carries over from the zero-skipping stuff below.
     */

    while (sq) {
      mp *y;
      y = mp_sqr(spare, x);
      y = mpmont_reduce(mm, y, y);
      spare = x; x = y;
      sq--;
    }

    /* --- Do the multiply --- */

    { mp *y = mpmont_mul(mm, spare, x, tab[i]); spare = x; x = y; }

    /* --- Now grind along through the rest of the bits --- */

    sq = z;
    for (;;) {
      if (!MP_RSTEP(&sc))
	goto done;
      if (MP_RBIT(&sc))
	break;
      sq++;
    }
  }

  /* --- Do a final round of squaring --- */

done:
  while (sq) {
    mp *y;
    y = mp_sqr(spare, x);
    y = mpmont_reduce(mm, y, y);
    spare = x; x = y;
    sq--;
  }

  /* --- Done --- */

  for (i = 0; i < TABSZ; i++)
    mp_drop(tab[i]);
  xfree(tab);
exit:
  mp_drop(d);
  mp_drop(spare);
  return (x);
}

/* --- @mpmont_exp@ --- *
 *
 * Arguments:	@mpmont *mm@ = pointer to Montgomery reduction context
 *		@mp *d@ = fake destination
 *		@mp *a@ = base
 *		@mp *e@ = exponent
 *
 * Returns:	Result, %$a^e \bmod m$%.
 */

mp *mpmont_exp(mpmont *mm, mp *d, mp *a, mp *e)
{
  d = mpmont_expr(mm, d, a, e);
  d = mpmont_reduce(mm, d, d);
  return (d);
}

/*----- Test rig ----------------------------------------------------------*/

#ifdef TEST_RIG

static int tcreate(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *mi = *(mp **)v[1].buf;
  mp *r = *(mp **)v[2].buf;
  mp *r2 = *(mp **)v[3].buf;

  mpmont mm;
  int ok = 1;

  mpmont_create(&mm, m);

  if (mm.mi->v[0] != mi->v[0]) {
    fprintf(stderr, "\n*** bad mi: found %lu, expected %lu",
	    (unsigned long)mm.mi->v[0], (unsigned long)mi->v[0]);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  if (!MP_EQ(mm.r, r)) {
    fputs("\n*** bad r", stderr);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\nexpected ", stderr); mp_writefile(r, stderr, 10);
    fputs("\n   found ", stderr); mp_writefile(mm.r, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  if (!MP_EQ(mm.r2, r2)) {
    fputs("\n*** bad r2", stderr);
    fputs("\nm = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\nexpected ", stderr); mp_writefile(r2, stderr, 10);
    fputs("\n   found ", stderr); mp_writefile(mm.r2, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  MP_DROP(m);
  MP_DROP(mi);
  MP_DROP(r);
  MP_DROP(r2);
  mpmont_destroy(&mm);
  assert(mparena_count(MPARENA_GLOBAL) == 0);
  return (ok);
}

static int tmul(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *a = *(mp **)v[1].buf;
  mp *b = *(mp **)v[2].buf;
  mp *r = *(mp **)v[3].buf;
  int ok = 1;

  mpmont mm;
  mpmont_create(&mm, m);

  {
    mp *qr = mp_mul(MP_NEW, a, b);
    mp_div(0, &qr, qr, m);

    if (!MP_EQ(qr, r)) {
      fputs("\n*** classical modmul failed", stderr);
      fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
      fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
      fputs("\n b = ", stderr); mp_writefile(b, stderr, 10);
      fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
      fputs("\nqr = ", stderr); mp_writefile(qr, stderr, 10);
      fputc('\n', stderr);
      ok = 0;
    }

    mp_drop(qr);
  }

  {
    mp *ar = mpmont_mul(&mm, MP_NEW, a, mm.r2);
    mp *br = mpmont_mul(&mm, MP_NEW, b, mm.r2);
    mp *mr = mpmont_mul(&mm, MP_NEW, ar, br);
    mr = mpmont_reduce(&mm, mr, mr);
    if (!MP_EQ(mr, r)) {
      fputs("\n*** montgomery modmul failed", stderr);
      fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
      fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
      fputs("\n b = ", stderr); mp_writefile(b, stderr, 10);
      fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
      fputs("\nmr = ", stderr); mp_writefile(mr, stderr, 10);
      fputc('\n', stderr);
      ok = 0;
    }
    MP_DROP(ar); MP_DROP(br);
    mp_drop(mr);
  }


  MP_DROP(m);
  MP_DROP(a);
  MP_DROP(b);
  MP_DROP(r);
  mpmont_destroy(&mm);
  assert(mparena_count(MPARENA_GLOBAL) == 0);
  return ok;
}

static int texp(dstr *v)
{
  mp *m = *(mp **)v[0].buf;
  mp *a = *(mp **)v[1].buf;
  mp *b = *(mp **)v[2].buf;
  mp *r = *(mp **)v[3].buf;
  mp *mr;
  int ok = 1;

  mpmont mm;
  mpmont_create(&mm, m);

  mr = mpmont_exp(&mm, MP_NEW, a, b);

  if (!MP_EQ(mr, r)) {
    fputs("\n*** montgomery modexp failed", stderr);
    fputs("\n m = ", stderr); mp_writefile(m, stderr, 10);
    fputs("\n a = ", stderr); mp_writefile(a, stderr, 10);
    fputs("\n e = ", stderr); mp_writefile(b, stderr, 10);
    fputs("\n r = ", stderr); mp_writefile(r, stderr, 10);
    fputs("\nmr = ", stderr); mp_writefile(mr, stderr, 10);
    fputc('\n', stderr);
    ok = 0;
  }

  MP_DROP(m);
  MP_DROP(a);
  MP_DROP(b);
  MP_DROP(r);
  MP_DROP(mr);
  mpmont_destroy(&mm);
  assert(mparena_count(MPARENA_GLOBAL) == 0);
  return ok;
}


static test_chunk tests[] = {
  { "create", tcreate, { &type_mp, &type_mp, &type_mp, &type_mp, 0 } },
  { "mul", tmul, { &type_mp, &type_mp, &type_mp, &type_mp, 0 } },
  { "exp", texp, { &type_mp, &type_mp, &type_mp, &type_mp, 0 } },
  { 0, 0, { 0 } },
};

int main(int argc, char *argv[])
{
  sub_init();
  test_run(argc, argv, tests, SRCDIR "/tests/mpmont");
  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
